import time

TIME_FMT = "%Y-%m-%d %H:%M:%S"


def get_time_str_by_now():
    time_stamp_now = time.time()
    return get_time_str_by_stamp(time_stamp_now)


def get_time_str_by_stamp(stamp: float):
    local_time = time.localtime(stamp)
    return time.strftime(TIME_FMT, local_time)



import asyncio
import random
from typing import Any, Callable, Coroutine, Literal, Optional, TypeVar, Union
from urllib.parse import quote, unquote

import aiohttp
from aiohttp import ClientSession

from yutto.exceptions import MaxRetryError
from yutto.utils.console.logger import Logger
from yutto.utils.file_buffer import AsyncFileBuffer

T = TypeVar("T")


class MaxRetry:
    def __init__(self, max_retry: int = 2):
        self.max_retry = max_retry

    def __call__(self, connect_once: Callable[..., Coroutine[Any, Any, T]]) -> Callable[..., Coroutine[Any, Any, T]]:
        async def connect_n_times(*args: Any, **kwargs: Any) -> T:
            retry = self.max_retry + 1
            while retry:
                try:
                    return await connect_once(*args, **kwargs)
                except (
                    aiohttp.client_exceptions.ClientPayloadError,  # type: ignore
                    aiohttp.client_exceptions.ServerDisconnectedError,  # type: ignore
                ):
                    await asyncio.sleep(0.5)
                    Logger.warning(f"抓取失败，正在重试，剩余 {retry - 1} 次")
                except asyncio.TimeoutError:
                    Logger.warning(f"抓取超时，正在重试，剩余 {retry - 1} 次")
                finally:
                    retry -= 1
            raise MaxRetryError("超出最大重试次数！")

        return connect_n_times


class Fetcher:
    proxy: Optional[str] = None
    trust_env: bool = True
    headers: dict[str, str] = {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692.71 Safari/537.36",
        "Referer": "https://www.bilibili.com",
    }
    cookies = {}
    semaphore: asyncio.Semaphore = asyncio.Semaphore(8)  # 初始使用较小的信号量用于抓取信息，下载时会重新设置一个较大的值

    @classmethod
    def set_proxy(cls, proxy: Union[Literal["no", "auto"], str]):
        if proxy == "auto":
            Fetcher.proxy = None
            Fetcher.trust_env = True
        elif proxy == "no":
            Fetcher.proxy = None
            Fetcher.trust_env = False
        else:
            Fetcher.proxy = proxy
            Fetcher.trust_env = False

    @classmethod
    def set_sessdata(cls, sessdata: str):
        # 先解码后编码是防止获取到的 SESSDATA 是已经解码后的（包含「,」）
        # 而番剧无法使用解码后的 SESSDATA
        Fetcher.cookies = {"SESSDATA": quote(unquote(sessdata))}

    @classmethod
    def set_semaphore(cls, num_workers: int):
        Fetcher.semaphore = asyncio.Semaphore(num_workers)

    @classmethod
    @MaxRetry(2)
    async def fetch_text(cls, session: ClientSession, url: str, encoding: Optional[str] = None) -> str:
        async with cls.semaphore:
            async with session.get(url, proxy=Fetcher.proxy) as resp:
                return await resp.text(encoding=encoding)

    @classmethod
    @MaxRetry(2)
    async def fetch_bin(cls, session: ClientSession, url: str) -> bytes:
        async with cls.semaphore:
            async with session.get(url, proxy=Fetcher.proxy) as resp:
                return await resp.read()

    @classmethod
    @MaxRetry(2)
    async def fetch_json(cls, session: ClientSession, url: str) -> Any:
        async with cls.semaphore:
            async with session.get(url, proxy=Fetcher.proxy) as resp:
                return await resp.json()

    @classmethod
    @MaxRetry(2)
    async def get_redirected_url(cls, session: ClientSession, url: str) -> str:
        async with session.get(
            url,
            proxy=Fetcher.proxy,
            ssl=False,
        ) as resp:
            async with cls.semaphore:
                return str(resp.url)

    @classmethod
    @MaxRetry(2)
    async def get_size(cls, session: ClientSession, url: str) -> Optional[int]:
        headers = session.headers.copy()
        headers["Range"] = "bytes=0-1"
        async with session.get(
            url,
            headers=headers,
            proxy=Fetcher.proxy,
            ssl=False,
        ) as resp:
            async with cls.semaphore:
                if resp.status == 206:
                    return int(resp.headers["Content-Range"].split("/")[-1])
                else:
                    return None

    @classmethod
    @MaxRetry(2)
    async def touch_url(cls, session: ClientSession, url: str):
        async with session.get(
            url,
            proxy=Fetcher.proxy,
            ssl=False,
        ) as resp:
            async with cls.semaphore:
                resp.close()

    @classmethod
    async def download_file_with_offset(
        cls,
        session: ClientSession,
        url: str,
        mirrors: list[str],
        file_buffer: AsyncFileBuffer,
        offset: int,
        size: Optional[int],
        stream: bool = True,
    ) -> None:
        async with cls.semaphore:
            done = False
            headers = session.headers.copy()
            url_pool = [url] + mirrors
            block_offset = 0
            while not done:
                try:
                    url = random.choice(url_pool)
                    headers["Range"] = "bytes={}-{}".format(
                        offset + block_offset, offset + size - 1 if size is not None else ""
                    )
                    async with session.get(
                        url,
                        headers=headers,
                        timeout=aiohttp.ClientTimeout(connect=5, sock_read=10),
                        proxy=Fetcher.proxy,
                        ssl=False,
                    ) as resp:
                        if stream:
                            while True:
                                # 如果直接用 1KiB 的话，会产生大量的块，消耗大量的 CPU 资源，
                                # 反而使得协程的优势不明显
                                # 而使用 1MiB 以上或者不使用流式下载方式时，由于分块太大，
                                # 导致进度条显示的实时速度并不准，波动太大，用户体验不佳，
                                # 因此取两者折中
                                chunk = await resp.content.read(2 ** 15)
                                if not chunk:
                                    break
                                await file_buffer.write(chunk, offset + block_offset)
                                block_offset += len(chunk)
                        else:
                            chunk = await resp.read()
                            await file_buffer.write(chunk, offset + block_offset)
                            block_offset += len(chunk)
                    # TODO: 是否需要校验总大小
                    done = True

                except (
                    aiohttp.client_exceptions.ClientPayloadError,  # type: ignore
                    aiohttp.client_exceptions.ServerDisconnectedError,  # type: ignore
                ):
                    await asyncio.sleep(0.5)
                    Logger.warning(f"文件 {file_buffer.file_path} 下载出错，尝试重新连接...")

                except asyncio.TimeoutError:
                    Logger.warning(f"文件 {file_buffer.file_path} 下载超时，尝试重新连接...")



import json
import re
from typing import Optional, TypedDict

from aiohttp import ClientSession

from yutto.api.info import get_video_info
from yutto.exceptions import NoAccessPermissionError, UnSupportedTypeError
from yutto.bilibili_typing.codec import audio_codec_map, video_codec_map
from yutto._typing import AudioUrlMeta, AvId, CId, MultiLangSubtitle, VideoUrlMeta
from yutto.utils.console.logger import Logger
from yutto.utils.fetcher import Fetcher
from yutto.utils.metadata import MetaData
from yutto.utils.time import get_time_str_by_now, get_time_str_by_stamp


class AcgVideoListItem(TypedDict):
    id: int
    name: str
    avid: AvId
    cid: CId
    metadata: Optional[MetaData]


async def get_acg_video_title(session: ClientSession, avid: AvId) -> str:
    return (await get_video_info(session, avid))["title"]


async def get_acg_video_pubdate(session: ClientSession, avid: AvId) -> str:
    return (await get_video_info(session, avid))["pubdate"]


async def get_acg_video_list(session: ClientSession, avid: AvId, with_metadata: bool = False) -> list[AcgVideoListItem]:
    list_api = "https://api.bilibili.com/x/player/pagelist?aid={aid}&bvid={bvid}&jsonp=jsonp"
    res_json = await Fetcher.fetch_json(session, list_api.format(**avid.to_dict()))
    if res_json.get("data") is None:
        Logger.warning(f"啊叻？视频 {avid} 不见了诶")
        return []
    acg_video_info: list[AcgVideoListItem] = [
        {
            "id": i + 1,
            "name": item["part"],
            "avid": avid,
            "cid": CId(str(item["cid"])),
            "metadata": None,
        }
        for i, item in enumerate(res_json["data"])
    ]

    if with_metadata:
        metadata_list = await get_acg_video_metadata(session, avid)
        assert len(metadata_list) == len(acg_video_info)
        for info, metadata in zip(acg_video_info, metadata_list):
            info["metadata"] = metadata
    return acg_video_info


async def get_acg_video_playurl(
    session: ClientSession, avid: AvId, cid: CId
) -> tuple[list[VideoUrlMeta], list[AudioUrlMeta]]:
    # 4048 = 16(useDash) | 64(useHDR) | 128(use4K) | 256(useDolby) | 512(useXXX) | 1024(use8K) | 2048(useAV1)
    play_api = "https://api.bilibili.com/x/player/playurl?avid={aid}&bvid={bvid}&cid={cid}&qn=127&type=&otype=json&fnver=0&fnval=4048&fourk=1"

    async with session.get(play_api.format(**avid.to_dict(), cid=cid), proxy=Fetcher.proxy) as resp:
        if not resp.ok:
            raise NoAccessPermissionError(f"无法下载该视频（cid: {cid}）")
        resp_json = await resp.json()
        if resp_json.get("data") is None:
            raise NoAccessPermissionError(f"无法下载该视频（cid: {cid}），原因：{resp_json.get('message')}")
        if resp_json["data"].get("dash") is None:
            raise UnSupportedTypeError(f"该视频（cid: {cid}）尚不支持 DASH 格式")
        # TODO: 处理 resp_json["data"]["dash"]["dolby"]，应当是 Dolby 的音频流
        return (
            [
                {
                    "url": video["base_url"],
                    "mirrors": video["backup_url"] if video["backup_url"] is not None else [],
                    "codec": video_codec_map[video["codecid"]],
                    "width": video["width"],
                    "height": video["height"],
                    "quality": video["id"],
                }
                for video in resp_json["data"]["dash"]["video"]
            ]
            if resp_json["data"]["dash"]["video"]
            else [],
            [
                {
                    "url": audio["base_url"],
                    "mirrors": audio["backup_url"] if audio["backup_url"] is not None else [],
                    "codec": audio_codec_map[audio["codecid"]],
                    "width": 0,
                    "height": 0,
                    "quality": audio["id"],
                }
                for audio in resp_json["data"]["dash"]["audio"]
            ]
            if resp_json["data"]["dash"]["audio"]
            else [],
        )


async def get_acg_video_subtitles(session: ClientSession, avid: AvId, cid: CId) -> list[MultiLangSubtitle]:
    subtitile_api = "https://api.bilibili.com/x/player.so?aid={aid}&bvid={bvid}&id=cid:{cid}"
    subtitile_url = subtitile_api.format(**avid.to_dict(), cid=cid)
    res_text = await Fetcher.fetch_text(session, subtitile_url)
    if subtitle_json_text_match := re.search(r"<subtitle>(.+)</subtitle>", res_text):
        subtitle_json = json.loads(subtitle_json_text_match.group(1))
        return [
            {
                "lang": sub_info["lan_doc"],
                "lines": (await Fetcher.fetch_json(session, "https:" + sub_info["subtitle_url"]))["body"],
            }
            for sub_info in subtitle_json["subtitles"]
        ]
    else:
        return []


async def get_acg_video_metadata(session: ClientSession, avid: AvId) -> list[MetaData]:
    web_interface_api = "https://api.bilibili.com/x/web-interface/view?aid={aid}&bvid={bvid}"
    web_interface_url = web_interface_api.format(**avid.to_dict())
    res_json = await Fetcher.fetch_json(session, web_interface_url)
    return [
        MetaData(
            title=page_info["part"],
            show_title=page_info["part"],
            plot=res_json["data"]["desc"],
            thumb=page_info["first_frame"] if page_info.get("first_frame") else res_json["data"]["pic"],
            premiered=get_time_str_by_stamp(res_json["data"]["pubdate"]),
            dataadded=get_time_str_by_now(),
            source="",  # TODO
            original_filename="",  # TODO
        )
        for page_info in res_json["data"]["pages"]
    ]



import asyncio
import platform
from typing import Any, Coroutine, TypeVar, Callable
from functools import wraps


from yutto.utils.console.logger import Logger

T = TypeVar("T")


def initial_async_policy():
    if platform.system() == "Windows":
        Logger.debug("Windows 平台，单独设置 EventLoopPolicy")
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


def install_uvloop():
    try:
        import uvloop
    except ImportError:
        Logger.warning("未安装 uvloop，无法使用其加速协程")
    else:
        uvloop.install()
        Logger.info("成功使用 uvloop 加速协程")


async def awaited_value(value: T) -> T:
    return value


def with_semaphore(
    func: Callable[..., Coroutine[Any, Any, T]], sem: asyncio.Semaphore
) -> Callable[..., Coroutine[Any, Any, T]]:
    @wraps(func)
    async def limited_func(*args: Any, **kwargs: Any) -> T:
        async with sem:
            return await func(*args, **kwargs)

    return limited_func



import argparse
import asyncio
import re
from typing import Optional

import aiohttp

from yutto._typing import AId, AvId, BvId, EpisodeData
from yutto.api.acg_video import get_acg_video_pubdate, get_acg_video_title
from yutto.exceptions import HttpStatusError, NoAccessPermissionError, NotFoundError, UnSupportedTypeError
from yutto.extractor._abc import SingleExtractor
from yutto.extractor.common import extract_acg_video_data
from yutto.utils.console.logger import Badge, Logger


class AcgVideoExtractor(SingleExtractor):
    """投稿视频单视频"""

    REGEX_AV = re.compile(r"https?://www\.bilibili\.com/video/av(?P<aid>\d+)(\?p=(?P<page>\d+))?")
    REGEX_BV = re.compile(r"https?://www\.bilibili\.com/video/(?P<bvid>(bv|BV)\w+)(\?p=(?P<page>\d+))?")

    REGEX_AV_ID = re.compile(r"av(?P<aid>\d+)(\?p=(?P<page>\d+))?")
    REGEX_BV_ID = re.compile(r"(?P<bvid>(bv|BV)\w+)(\?p=(?P<page>\d+))?")

    page: int
    avid: AvId

    def resolve_shortcut(self, id: str) -> tuple[bool, str]:
        matched = False
        url = id
        if match_obj := self.REGEX_AV_ID.match(id):
            page: int = 1
            if match_obj.group("page") is not None:
                page = int(match_obj.group("page"))
            url = f"https://www.bilibili.com/video/av{match_obj.group('aid')}?p={page}"
            matched = True
        elif match_obj := self.REGEX_BV_ID.match(id):
            page: int = 1
            if match_obj.group("page") is not None:
                page = int(match_obj.group("page"))
            url = f"https://www.bilibili.com/video/{match_obj.group('bvid')}?p={page}"
            matched = True
        return matched, url

    def match(self, url: str) -> bool:
        if (match_obj := self.REGEX_AV.match(url)) or (match_obj := self.REGEX_BV.match(url)):
            self.page: int = 1
            if "aid" in match_obj.groupdict().keys():
                self.avid = AId(match_obj.group("aid"))
            else:
                self.avid = BvId(match_obj.group("bvid"))
            if match_obj.group("page") is not None:
                self.page = int(match_obj.group("page"))
            return True
        else:
            return False

    async def extract(self, session: aiohttp.ClientSession, args: argparse.Namespace) -> Optional[EpisodeData]:
        title, pubdate = await asyncio.gather(
            get_acg_video_title(session, self.avid),
            get_acg_video_pubdate(session, self.avid),
        )
        Logger.custom(title, Badge("投稿视频", fore="black", back="cyan"))
        try:
            return await extract_acg_video_data(
                session,
                self.avid,
                self.page,
                None,
                args,
                {"title": title, "pubdate": pubdate},
                "{title}",
            )
        except (NoAccessPermissionError, HttpStatusError, UnSupportedTypeError, NotFoundError) as e:
            Logger.error(e.message)
            return None



import argparse
import os
import sys
from typing import Optional

import aiohttp

from yutto.api.acg_video import AcgVideoListItem, get_acg_video_list, get_acg_video_playurl, get_acg_video_subtitles
from yutto.api.bangumi import (
    BangumiListItem,
    get_bangumi_list,
    get_bangumi_playurl,
    get_bangumi_subtitles,
    get_season_id_by_episode_id,
)
from yutto.api.danmaku import get_danmaku
from yutto.exceptions import ErrorCode
from yutto.processor.path_resolver import UNKNOWN, PathTemplateVariableDict, resolve_path_template
from yutto._typing import AvId, EpisodeData, EpisodeId
from yutto.utils.console.logger import Logger
from yutto.utils.danmaku import EmptyDanmakuData


async def extract_bangumi_data(
    session: aiohttp.ClientSession,
    episode_id: EpisodeId,
    bangumi_info: Optional[BangumiListItem],
    args: argparse.Namespace,
    subpath_variables: PathTemplateVariableDict,
    auto_subpath_template: str = "{name}",
) -> EpisodeData:
    season_id = await get_season_id_by_episode_id(session, episode_id)
    # 如果不包含详细信息，需从列表中解析
    if bangumi_info is None:
        bangumi_list = await get_bangumi_list(session, season_id, with_metadata=args.with_metadata)
        for bangumi_item in bangumi_list:
            if bangumi_item["episode_id"] == episode_id:
                bangumi_info = bangumi_item
                break
        else:
            Logger.error("在列表中未找到该剧集")
            sys.exit(ErrorCode.EPISODE_NOT_FOUND_ERROR.value)
    avid = bangumi_info["avid"]
    cid = bangumi_info["cid"]
    name = bangumi_info["name"]
    id = bangumi_info["id"]
    videos, audios = await get_bangumi_playurl(session, avid, episode_id, cid)
    subtitles = await get_bangumi_subtitles(session, avid, cid) if not args.no_subtitle else []
    danmaku = await get_danmaku(session, cid, args.danmaku_format) if not args.no_danmaku else EmptyDanmakuData
    metadata = bangumi_info["metadata"]
    subpath_variables_base: PathTemplateVariableDict = {
        "id": id,
        "name": name,
        "title": UNKNOWN,
        "username": UNKNOWN,
        "series_title": UNKNOWN,
        "pubdate": UNKNOWN,
    }
    subpath_variables_base.update(subpath_variables)
    subpath = resolve_path_template(args.subpath_template, auto_subpath_template, subpath_variables_base)
    output_dir, filename = os.path.split(os.path.join(args.dir, subpath))
    return EpisodeData(
        videos=videos,
        audios=audios,
        subtitles=subtitles,
        danmaku=danmaku,
        metadata=metadata,
        output_dir=output_dir,
        tmp_dir=args.tmp_dir or output_dir,
        filename=filename,
    )


async def extract_acg_video_data(
    session: aiohttp.ClientSession,
    avid: AvId,
    page: int,
    acg_video_info: Optional[AcgVideoListItem],
    args: argparse.Namespace,
    subpath_variables: PathTemplateVariableDict,
    auto_subpath_template: str = "{title}",
) -> EpisodeData:
    acg_video_list = await get_acg_video_list(session, avid, with_metadata=args.with_metadata)
    if acg_video_info is None:
        acg_video_info = acg_video_list[page - 1]
    cid = acg_video_info["cid"]
    name = acg_video_info["name"]
    id = acg_video_info["id"]
    videos, audios = await get_acg_video_playurl(session, avid, cid)
    subtitles = await get_acg_video_subtitles(session, avid, cid) if not args.no_subtitle else []
    danmaku = await get_danmaku(session, cid, args.danmaku_format) if not args.no_danmaku else EmptyDanmakuData
    metadata = acg_video_info["metadata"]
    subpath_variables_base: PathTemplateVariableDict = {
        "id": id,
        "name": name,
        "title": UNKNOWN,
        "username": UNKNOWN,
        "series_title": UNKNOWN,
        "pubdate": UNKNOWN,
    }
    subpath_variables_base.update(subpath_variables)
    subpath = resolve_path_template(args.subpath_template, auto_subpath_template, subpath_variables_base)
    output_dir, filename = os.path.split(os.path.join(args.dir, subpath))
    return EpisodeData(
        videos=videos,
        audios=audios,
        subtitles=subtitles,
        danmaku=danmaku,
        metadata=metadata,
        output_dir=output_dir,
        tmp_dir=args.tmp_dir or output_dir,
        filename=filename,
    )



import argparse
import asyncio
import re
from typing import Any, Coroutine, Optional

import aiohttp

from yutto._typing import EpisodeData, MId, SeriesId
from yutto.api.acg_video import AcgVideoListItem, get_acg_video_list, get_acg_video_pubdate, get_acg_video_title
from yutto.api.space import (
    get_collection_avids,
    get_collection_title,
    get_medialist_avids,
    get_medialist_title,
    get_uploader_name,
)
from yutto.exceptions import HttpStatusError, NoAccessPermissionError, NotFoundError, UnSupportedTypeError
from yutto.extractor._abc import BatchExtractor
from yutto.extractor.common import extract_acg_video_data
from yutto.utils.console.logger import Badge, Logger
from yutto.utils.fetcher import Fetcher


class SeriesExtractor(BatchExtractor):
    """视频合集和视频列表"""

    REGEX_SERIES = re.compile(
        r"https?://space\.bilibili\.com/(?P<mid>\d+)/channel/seriesdetail\?sid=(?P<series_id>\d+)"
    )
    REGEX_MEDIA_LIST = re.compile(
        r"https?://www\.bilibili\.com/medialist/play/(?P<mid>\d+)\?business=space_series&business_id=(?P<series_id>\d+)"
    )
    REGEX_COLLECTIOMS = re.compile(
        r"https?://space\.bilibili\.com/(?P<mid>\d+)/channel/collectiondetail\?sid=(?P<series_id>\d+)"
    )

    mid: MId
    series_id: SeriesId
    is_collection: bool

    def match(self, url: str) -> bool:
        if (
            (match_obj := self.REGEX_MEDIA_LIST.match(url))
            or (match_obj := self.REGEX_SERIES.match(url))
            or (match_obj := self.REGEX_COLLECTIOMS.match(url))
        ):
            self.mid = MId(match_obj.group("mid"))
            self.series_id = SeriesId(match_obj.group("series_id"))
            self.is_collection = True if self.REGEX_COLLECTIOMS.match(url) else False
            return True
        else:
            return False

    async def extract(
        self, session: aiohttp.ClientSession, args: argparse.Namespace
    ) -> list[Coroutine[Any, Any, Optional[tuple[int, EpisodeData]]]]:
        # 视频合集
        if self.is_collection:
            username, series_title = await asyncio.gather(
                get_uploader_name(session, self.mid), get_collection_title(session, self.series_id)
            )
            Logger.custom(series_title, Badge("视频合集", fore="black", back="cyan"))

            acg_video_list = [
                acg_video_item
                for avid in await get_collection_avids(session, self.series_id)
                for acg_video_item in await get_acg_video_list(session, avid, with_metadata=args.with_metadata)
            ]
        # 视频列表
        else:
            username, series_title = await asyncio.gather(
                get_uploader_name(session, self.mid), get_medialist_title(session, self.series_id)
            )
            Logger.custom(series_title, Badge("视频列表", fore="black", back="cyan"))

            acg_video_list = [
                acg_video_item
                for avid in await get_medialist_avids(session, self.series_id)
                for acg_video_item in await get_acg_video_list(session, avid, with_metadata=args.with_metadata)
            ]

        return [
            self._parse_episodes_data(
                session,
                args,
                series_title,
                username,
                i,
                acg_video_item,
            )
            for i, acg_video_item in enumerate(acg_video_list)
        ]

    async def _parse_episodes_data(
        self,
        session: aiohttp.ClientSession,
        args: argparse.Namespace,
        series_title: str,
        username: str,
        i: int,
        acg_video_item: AcgVideoListItem,
    ) -> Optional[tuple[int, EpisodeData]]:
        try:
            _, title, pubdate = await asyncio.gather(
                Fetcher.touch_url(session, acg_video_item["avid"].to_url()),
                get_acg_video_title(session, acg_video_item["avid"]),
                get_acg_video_pubdate(session, acg_video_item["avid"]),
            )
            return (
                i,
                await extract_acg_video_data(
                    session,
                    acg_video_item["avid"],
                    i + 1,
                    acg_video_item,
                    args,
                    {
                        "series_title": series_title,
                        "username": username,  # 虽然默认模板的用不上，但这里可以提供一下
                        "title": title,
                        "pubdate": pubdate,
                    },
                    "{series_title}/{title}/{name}",
                ),
            )
        except (NoAccessPermissionError, HttpStatusError, UnSupportedTypeError, NotFoundError) as e:
            Logger.error(e.message)
            return None



from typing import TypeVar, Optional, Iterable

T = TypeVar("T")


def filter_none_value(l: Iterable[Optional[T]]) -> Iterable[T]:
    result: Iterable[T] = []
    for item in l:
        if item is not None:
            result.append(item)
    return result



import argparse
import asyncio
from typing import Any, Coroutine, Optional

import aiohttp

from yutto._typing import EpisodeData
from yutto.utils.console.logger import Logger


class Extractor:
    def resolve_shortcut(self, id: str) -> tuple[bool, str]:
        matched = False
        url = id
        return (matched, url)

    def match(self, url: str) -> bool:
        raise NotImplementedError

    async def __call__(self, session: aiohttp.ClientSession, args: argparse.Namespace) -> list[EpisodeData]:
        raise NotImplementedError


class SingleExtractor(Extractor):
    async def __call__(self, session: aiohttp.ClientSession, args: argparse.Namespace) -> list[EpisodeData]:
        episode_data = await self.extract(session, args)
        if episode_data is not None:
            return [episode_data]
        return []

    async def extract(self, session: aiohttp.ClientSession, args: argparse.Namespace) -> Optional[EpisodeData]:
        raise NotImplementedError


class BatchExtractor(Extractor):
    async def __call__(self, session: aiohttp.ClientSession, args: argparse.Namespace) -> list[EpisodeData]:
        download_list: list[tuple[int, EpisodeData]] = []
        coroutine_list = await self.extract(session, args)
        num_videos = len(coroutine_list)
        # 先解析各种资源链接
        for i, coro in enumerate(asyncio.as_completed(coroutine_list)):
            Logger.status.set(f"正在努力解析第 {i+1}/{num_videos} 个视频")
            results = await coro
            if results is not None:
                download_list.append(results)

        # 由于 asyncio.as_completed 的顺序是按照完成顺序的，所以需要重新排序下
        download_list.sort(key=lambda x: x[0])

        return [item for _, item in download_list]

    async def extract(
        self, session: aiohttp.ClientSession, args: argparse.Namespace
    ) -> list[Coroutine[Any, Any, Optional[tuple[int, EpisodeData]]]]:
        raise NotImplementedError



import argparse
import asyncio
import re
from typing import Any, Coroutine, Optional

import aiohttp

from yutto._typing import AId, AvId, BvId, EpisodeData
from yutto.api.acg_video import AcgVideoListItem, get_acg_video_list, get_acg_video_pubdate, get_acg_video_title
from yutto.exceptions import HttpStatusError, NoAccessPermissionError, UnSupportedTypeError
from yutto.extractor._abc import BatchExtractor
from yutto.extractor.common import extract_acg_video_data
from yutto.processor.filter import parse_episodes
from yutto.utils.console.logger import Badge, Logger


class AcgVideoBatchExtractor(BatchExtractor):
    """投稿视频批下载"""

    REGEX_AV = re.compile(r"https?://www\.bilibili\.com/video/av(?P<aid>\d+)(\?p=(?P<page>\d+))?")
    REGEX_BV = re.compile(r"https?://www\.bilibili\.com/video/(?P<bvid>(bv|BV)\w+)(\?p=(?P<page>\d+))?")

    REGEX_AV_ID = re.compile(r"av(?P<aid>\d+)(\?p=(?P<page>\d+))?")
    REGEX_BV_ID = re.compile(r"(?P<bvid>(bv|BV)\w+)(\?p=(?P<page>\d+))?")

    avid: AvId

    def resolve_shortcut(self, id: str) -> tuple[bool, str]:
        matched = False
        url = id
        if match_obj := self.REGEX_AV_ID.match(id):
            page: int = 1
            if match_obj.group("page") is not None:
                page = int(match_obj.group("page"))
            url = f"https://www.bilibili.com/video/av{match_obj.group('aid')}?p={page}"
            matched = True
        elif match_obj := self.REGEX_BV_ID.match(id):
            page: int = 1
            if match_obj.group("page") is not None:
                page = int(match_obj.group("page"))
            url = f"https://www.bilibili.com/video/{match_obj.group('bvid')}?p={page}"
            matched = True
        return matched, url

    def match(self, url: str) -> bool:
        if (match_obj := self.REGEX_AV.match(url)) or (match_obj := self.REGEX_BV.match(url)):
            if "aid" in match_obj.groupdict().keys():
                self.avid = AId(match_obj.group("aid"))
            else:
                self.avid = BvId(match_obj.group("bvid"))
            return True
        else:
            return False

    async def extract(
        self, session: aiohttp.ClientSession, args: argparse.Namespace
    ) -> list[Coroutine[Any, Any, Optional[tuple[int, EpisodeData]]]]:
        title, pubdate, acg_video_list = await asyncio.gather(
            get_acg_video_title(session, self.avid),
            get_acg_video_pubdate(session, self.avid),
            get_acg_video_list(session, self.avid, with_metadata=args.with_metadata),
        )
        Logger.custom(title, Badge("投稿视频", fore="black", back="cyan"))

        # 选集过滤
        episodes = parse_episodes(args.episodes, len(acg_video_list))
        acg_video_list = list(filter(lambda item: item["id"] in episodes, acg_video_list))

        return [
            self._parse_episodes_data(
                session,
                self.avid,
                args,
                title,
                pubdate,
                i,
                acg_video_item,
            )
            for i, acg_video_item in enumerate(acg_video_list)
        ]

    async def _parse_episodes_data(
        self,
        session: aiohttp.ClientSession,
        avid: AvId,
        args: argparse.Namespace,
        title: str,
        pubdate: str,
        i: int,
        acg_video_item: AcgVideoListItem,
    ) -> Optional[tuple[int, EpisodeData]]:
        try:
            return (
                i,
                await extract_acg_video_data(
                    session,
                    avid,
                    i + 1,
                    acg_video_item,
                    args,
                    {"title": title, "pubdate": pubdate},
                    "{title}/{name}",
                ),
            )
        except (NoAccessPermissionError, HttpStatusError, UnSupportedTypeError) as e:
            Logger.error(e.message)
            return None



import argparse
import asyncio
import re
from typing import Any, Coroutine, Optional

import aiohttp

from yutto._typing import EpisodeData, MId
from yutto.api.acg_video import AcgVideoListItem, get_acg_video_list, get_acg_video_pubdate, get_acg_video_title
from yutto.api.space import get_all_favourites, get_favourite_avids, get_uploader_name
from yutto.exceptions import HttpStatusError, NoAccessPermissionError, NotFoundError, UnSupportedTypeError
from yutto.extractor._abc import BatchExtractor
from yutto.extractor.common import extract_acg_video_data
from yutto.utils.console.logger import Badge, Logger
from yutto.utils.fetcher import Fetcher


class FavouritesAllExtractor(BatchExtractor):
    """用户单一收藏夹"""

    REGEX_FAV_ALL = re.compile(r"https?://space\.bilibili\.com/(?P<mid>\d+)/favlist\?fid=(?P<fid>\d+)")

    mid: MId

    def match(self, url: str) -> bool:
        if match_obj := self.REGEX_FAV_ALL.match(url):
            self.mid = MId(match_obj.group("mid"))
            return True
        else:
            return False

    async def extract(
        self, session: aiohttp.ClientSession, args: argparse.Namespace
    ) -> list[Coroutine[Any, Any, Optional[tuple[int, EpisodeData]]]]:
        username = await get_uploader_name(session, self.mid)
        Logger.custom(username, Badge("用户收藏夹", fore="black", back="cyan"))

        acg_video_list = [
            (acg_video_item, fav["title"])
            for fav in await get_all_favourites(session, self.mid)
            for avid in await get_favourite_avids(session, fav["fid"])
            for acg_video_item in await get_acg_video_list(session, avid, with_metadata=args.with_metadata)
        ]

        return [
            self._parse_episodes_data(
                session,
                args,
                username,
                series_title,
                i,
                acg_video_item,
            )
            for i, (acg_video_item, series_title) in enumerate(acg_video_list)
        ]

    async def _parse_episodes_data(
        self,
        session: aiohttp.ClientSession,
        args: argparse.Namespace,
        username: str,
        series_title: str,
        i: int,
        acg_video_item: AcgVideoListItem,
    ) -> Optional[tuple[int, EpisodeData]]:
        pubdate = await get_acg_video_pubdate(session, acg_video_item["avid"])
        try:
            _, title = await asyncio.gather(
                Fetcher.touch_url(session, acg_video_item["avid"].to_url()),
                get_acg_video_title(session, acg_video_item["avid"]),
            )
            return (
                i,
                await extract_acg_video_data(
                    session,
                    acg_video_item["avid"],
                    i + 1,
                    acg_video_item,
                    args,
                    {
                        "title": title,
                        "username": username,
                        "series_title": series_title,
                        "pubdate": pubdate,
                    },
                    "{username}的收藏夹/{series_title}/{title}/{name}",
                ),
            )
        except (NoAccessPermissionError, HttpStatusError, UnSupportedTypeError, NotFoundError) as e:
            Logger.error(e.message)
            return None



from typing import Any


class aobject(object):
    """Inheriting this class allows you to define an async __ainit__.

    So you can create objects by doing something like `await MyClass(params)`
    Based on : https://stackoverflow.com/questions/33128325/how-to-set-class-attribute-with-await-in-init
    """

    async def __new__(cls, *args: Any, **kwargs: Any):
        instance = super().__new__(cls)
        await instance.__ainit__(*args, **kwargs)
        return instance

    async def __ainit__(self, *args: Any, **kwargs: Any):
        pass



import argparse
import asyncio
import re
from typing import Any, Coroutine, Optional

import aiohttp

from yutto._typing import EpisodeData, FavouriteMetaData, FId, MId
from yutto.api.acg_video import AcgVideoListItem, get_acg_video_list, get_acg_video_pubdate, get_acg_video_title
from yutto.api.space import get_favourite_avids, get_favourite_info, get_uploader_name
from yutto.exceptions import HttpStatusError, NoAccessPermissionError, NotFoundError, UnSupportedTypeError
from yutto.extractor._abc import BatchExtractor
from yutto.extractor.common import extract_acg_video_data
from yutto.processor.filter import parse_episodes
from yutto.utils.console.logger import Badge, Logger
from yutto.utils.fetcher import Fetcher


class FavouritesExtractor(BatchExtractor):
    """用户所有收藏夹"""

    REGEX_FAV = re.compile(r"https?://space\.bilibili\.com/(?P<mid>\d+)/favlist\?fid=(?P<fid>\d+)")

    mid: MId
    fid: FId

    def match(self, url: str) -> bool:
        if match_obj := self.REGEX_FAV.match(url):
            self.mid = MId(match_obj.group("mid"))
            self.fid = FId(match_obj.group("fid"))
            return True
        else:
            return False

    async def extract(
        self, session: aiohttp.ClientSession, args: argparse.Namespace
    ) -> list[Coroutine[Any, Any, Optional[tuple[int, EpisodeData]]]]:
        username, favourite_info = await asyncio.gather(
            get_uploader_name(session, self.mid),
            get_favourite_info(session, self.fid),
        )
        Logger.custom(favourite_info["title"], Badge("收藏夹", fore="black", back="cyan"))

        acg_video_list = [
            acg_video_item
            for avid in await get_favourite_avids(session, self.fid)
            for acg_video_item in await get_acg_video_list(session, avid, with_metadata=args.with_metadata)
        ]
        # 选集过滤
        episodes = parse_episodes(args.episodes, len(acg_video_list))
        acg_video_list = list(filter(lambda item: item["id"] in episodes, acg_video_list))

        return [
            self._parse_episodes_data(
                session,
                args,
                username,
                favourite_info,
                i,
                acg_video_item,
            )
            for i, acg_video_item in enumerate(acg_video_list)
        ]

    async def _parse_episodes_data(
        self,
        session: aiohttp.ClientSession,
        args: argparse.Namespace,
        username: str,
        favourite_info: FavouriteMetaData,
        i: int,
        acg_video_item: AcgVideoListItem,
    ) -> Optional[tuple[int, EpisodeData]]:
        try:
            # 在使用 SESSDATA 时，如果不去事先 touch 一下视频链接的话，是无法获取 episode_data 的
            # 至于为什么前面那俩（投稿视频页和番剧页）不需要额外 touch，因为在 get_redirected_url 阶段连接过了呀
            _, title, pubdate = await asyncio.gather(
                Fetcher.touch_url(session, acg_video_item["avid"].to_url()),
                get_acg_video_title(session, acg_video_item["avid"]),
                get_acg_video_pubdate(session, acg_video_item["avid"]),
            )
            return (
                i,
                await extract_acg_video_data(
                    session,
                    acg_video_item["avid"],
                    i + 1,
                    acg_video_item,
                    args,
                    {
                        "title": title,
                        "username": username,
                        "series_title": favourite_info["title"],
                        "pubdate": pubdate,
                    },
                    "{username}的收藏夹/{series_title}/{title}/{name}",
                ),
            )
        except (NoAccessPermissionError, HttpStatusError, UnSupportedTypeError, NotFoundError) as e:
            Logger.error(e.message)
            return None



from .acg_video import AcgVideoExtractor
from .acg_video_batch import AcgVideoBatchExtractor
from .bangumi import BangumiExtractor
from .bangumi_batch import BangumiBatchExtractor
from .uploader_all_videos import UploaderAllVideosExtractor
from .favourites import FavouritesExtractor
from .favourites_all import FavouritesAllExtractor
from .series import SeriesExtractor

__all__ = [
    "AcgVideoExtractor",
    "AcgVideoBatchExtractor",
    "BangumiExtractor",
    "BangumiBatchExtractor",
    "UploaderAllVideosExtractor",
    "FavouritesExtractor",
    "FavouritesAllExtractor",
    "SeriesExtractor",
]
